import { Action } from '../display/action.js';

class ContractRecord {
  constructor(id, instance) {
    this.id = id;
    this.instance = instance;
  }
}

export class VMHandle {
  constructor(vm, sender) {
    this.vm = vm;
    this.sender = sender;
    this.actions = {
      call: new Action('call', vm.scene),
      return: new Action('return', vm.scene),
    };
  }

  /**
   * @param {string} id Contract ID
   * @param {string} method
   */
  async callContract(id, method, ...args) {
    const instance = this.vm.getContractInstance(id);
    const fn = instance[method];
    if (!fn) throw new Error(`Contract ${id} method ${method} not found!`);
    await this.actions.call.log(this.sender, instance, method);
    const result = await fn.call(instance, this.sender, ...args);
    await this.actions.return.log(instance, this.sender, undefined, undefined, '-->>');
    return result;
  }
}

export class VM {
  constructor(scene) {
    this.scene = scene;
    this.contracts = new Map();
  }

  /**
   * @param {string} id
   * @param {class} ContractClass
   * @param {any[]} ...args Passed to contractClass constructor after `vm`
   */
  addContract(id, ContractClass, ...args) {
    const instance = new ContractClass(this, ...args);
    const contract = new ContractRecord(id, instance);
    this.contracts.set(id, contract);
  }

  getHandle(sender) {
    return new VMHandle(this, sender);
  }

  /**
   * @param {string} id
   */
  getContract(id) {
    return this.contracts.get(id);
  }

  /**
   * @param {string} id
   */
  getContractInstance(id) {
    return this.getContract(id)?.instance;
  }
}
